from dataclasses import dataclass
from logging import getLogger
from typing import Optional, Set

from django.conf import settings
from identitylib.identifiers import Identifier
from rest_framework import authentication
from rest_framework.exceptions import AuthenticationFailed
from rest_framework.request import Request

from .id_token import InvalidIdTokenError, verify_id_token_for_api_backend

LOG = getLogger(__name__)


@dataclass(eq=True)
class APIGatewayAuthenticationDetails:
    """
    A dataclass representing the authentication information passed from the API Gateway.

    """

    principal_identifier: Identifier
    scopes: Set[str]
    app_id: Optional[str] = None
    client_id: Optional[str] = None


class APIGatewayAuthentication(authentication.BaseAuthentication):
    """
    An Authentication provider which interprets the headers provided by the API Gateway.

    This library expects to only be used within an application that is deployed behind and can
    only be invoked by the API Gateway, and therefore relies on the fact that the headers
    provided are authoritative.

    """

    def authenticate(self, request: Request):
        # This import is needed here because APIGatewayUser references
        # AnonymousUser from django.contrib.auth which, in turn, means that
        # applications have to be ready at import time.
        #
        # This file is imported from apigatewayauth/__init__.py and
        # apigatewayauth is imported at application configure time. The net
        # upshot is that we cannot import directly or indirectly from
        # django.contrib.auth at the top of this file and have to do it here.
        #
        # We can't remove the imports at the top of apigatewayauth/__init__.py
        # because we have users of this library which set the DRF default
        # authentication class to "apigatewayauth.APIGatewayAuthentication".
        #
        # These users need to be fixed up to use
        # "apigatewayauth.authentication.APIGatewayAuthentication" instead and
        # then we can move this import back where it belongs.
        from .user import APIGatewayUser

        if not request.META.get("HTTP_X_API_ORG_NAME", None):
            # bail early if we look like we're not being called by the API Gateway
            return None

        try:
            # We should have "Bearer ..." in the authorization header.
            auth_header = request.headers.get("Authorization", "")
            if not auth_header.startswith("Bearer "):
                raise AuthenticationFailed("Bearer token not present")

            # Extract the id token from the API Gateway for verification.
            _, token = auth_header.split(" ")
            expected_audience = getattr(
                settings, "API_GATEWAY_JWT_EXPECTED_AUDIENCE", request.build_absolute_uri("/")
            )
            trusted_issuers = getattr(settings, "API_GATEWAY_JWT_TRUSTED_ISSUERS", None)
            expected_authorised_parties = getattr(
                settings, "API_GATEWAY_JWT_EXPECTED_AUTHORISED_PARTIES", None
            )
            try:
                verify_id_token_for_api_backend(
                    token,
                    expected_audience,
                    certs_url=getattr(settings, "API_GATEWAY_JWT_ISSUER_CERTS_URL", None),
                    trusted_issuers=trusted_issuers,
                    expected_authorised_parties=expected_authorised_parties,
                )
            except InvalidIdTokenError as e:
                LOG.info(f"Incoming API token failed verification: {e}")
                raise AuthenticationFailed("Invalid API Gateway token") from e
        except AuthenticationFailed as e:
            if getattr(settings, "API_GATEWAY_ENFORCE_ID_TOKEN_VERIFICATION", False):
                raise e
            else:
                LOG.warning(
                    "API_GATEWAY_ENFORCE_ID_TOKEN_VERIFICATION is False. "
                    f"Allowing incoming request with invalid authentication: {e}"
                )

        if not request.META.get("HTTP_X_API_OAUTH2_USER", None):
            raise AuthenticationFailed("Could not authenticate using x-api-* headers")

        try:
            principal_identifier = Identifier.from_string(
                request.META["HTTP_X_API_OAUTH2_USER"], find_by_alias=True
            )
        except Exception:
            raise AuthenticationFailed("Invalid principal identifier")

        auth = APIGatewayAuthenticationDetails(
            principal_identifier=principal_identifier,
            scopes=set(filter(bool, request.META.get("HTTP_X_API_OAUTH2_SCOPE", "").split(" "))),
            # the following will only be populated for confidential clients
            app_id=request.META.get("HTTP_X_API_DEVELOPER_APP_ID", None),
            client_id=request.META.get("HTTP_X_API_OAUTH2_CLIENT_ID", None),
        )
        user = APIGatewayUser(auth)
        return user, auth



from django.contrib.auth.models import AnonymousUser

from .authentication import APIGatewayAuthenticationDetails


class APIGatewayUser(AnonymousUser):
    """
    A Django user representing the authenticated principal. This user is not
    backed by a database object and so they can have no permissions in the
    Django sense.
    """

    def __init__(self, auth: APIGatewayAuthenticationDetails):
        super().__init__()
        self.username = self.id = self.pk = str(auth.principal_identifier)

    @property
    def is_anonymous(self):
        return False

    @property
    def is_authenticated(self):
        return True

    def __str__(self):
        return self.username



import secrets
from unittest import mock

from django.test import TestCase, override_settings
from google.auth.exceptions import GoogleAuthError
from identitylib.identifiers import Identifier, IdentifierSchemes
from rest_framework.exceptions import AuthenticationFailed
from rest_framework.test import APIRequestFactory

from apigatewayauth.authentication import (
    APIGatewayAuthentication,
    APIGatewayAuthenticationDetails,
)


@override_settings(API_GATEWAY_ENFORCE_ID_TOKEN_VERIFICATION=True)
class APIGatewayAuthTestCase(TestCase):
    def setUp(self):
        super().setUp()

        self.request_factory = APIRequestFactory()
        self.auth = APIGatewayAuthentication()

        # Patch the token verification function. The default side effect is to check that the token
        # matches self.expected_token.
        self.expected_token = secrets.token_urlsafe()

        def verify_token(token: str, *args, **kwargs):
            if token != self.expected_token:
                raise GoogleAuthError("Token did not match mock token")
            return {
                "iss": "https://accounts.google.com",
                "azp": "api-gateway@api-meta-2555105a.iam.gserviceaccount.com",
            }

        verify_token_patcher = mock.patch(
            "google.oauth2.id_token.verify_token", side_effect=verify_token
        )
        self.mock_verify_token = verify_token_patcher.start()
        self.addCleanup(verify_token_patcher.stop)

    def request_with_headers(self, headers={}, *, include_authorization=True):
        parsed_headers = {
            f'HTTP_{key.upper().replace("-", "_")}': value for key, value in headers.items()
        }
        if include_authorization:
            parsed_headers["HTTP_AUTHORIZATION"] = f"Bearer {self.expected_token}"
        return self.request_factory.get("/", **parsed_headers)

    def test_bails_early_without_api_org(self):
        self.assertIsNone(
            self.auth.authenticate(self.request_with_headers({"Accept": "application/json"}))
        )

    def test_throws_without_auth_details(self):
        with self.assertRaisesMessage(
            AuthenticationFailed, "Could not authenticate using x-api-* headers"
        ):
            self.auth.authenticate(self.request_with_headers({"x-api-org-name": "test"}))

    def test_throws_without_principal_identifier(self):
        with self.assertRaisesMessage(
            AuthenticationFailed, "Could not authenticate using x-api-* headers"
        ):
            self.auth.authenticate(
                self.request_with_headers(
                    {"x-api-org-name": "test", "x-api-developer-app-class": "public"}
                )
            )

    def test_throws_with_bad_principal_identifier(self):
        with self.assertRaisesMessage(AuthenticationFailed, "Invalid principal identifier"):
            self.auth.authenticate(
                self.request_with_headers(
                    {
                        "x-api-org-name": "test",
                        "x-api-developer-app-class": "public",
                        "x-api-oauth2-user": "Monty Dawson",
                    }
                )
            )

    def test_can_use_any_identifier_scheme_in_principal_identifier(self):
        for scheme in IdentifierSchemes.get_registered_schemes():
            _, auth = self.auth.authenticate(
                self.request_with_headers(
                    {
                        "x-api-org-name": "test",
                        "x-api-developer-app-class": "public",
                        "x-api-oauth2-user": str(Identifier("1000", scheme)),
                    }
                )
            )
            self.assertEqual(auth.principal_identifier, Identifier("1000", scheme))

    def test_throws_with_unknown_identifier_type(self):
        with self.assertRaisesMessage(AuthenticationFailed, "Invalid principal identifier"):
            self.auth.authenticate(
                self.request_with_headers(
                    {
                        "x-api-org-name": "test",
                        "x-api-developer-app-class": "public",
                        "x-api-oauth2-user": "wgd23@gmail.com",
                    }
                )
            )

    def test_returns_client_details_for_valid_auth(self):
        user, auth = self.auth.authenticate(
            self.request_with_headers(
                {
                    "x-api-org-name": "test",
                    "x-api-developer-app-class": "public",
                    "x-api-oauth2-user": str(Identifier("a123", IdentifierSchemes.CRSID)),
                }
            )
        )
        self.assertEqual(user.id, str(Identifier("a123", IdentifierSchemes.CRSID)))

        self.assertEqual(
            auth,
            APIGatewayAuthenticationDetails(
                Identifier("a123", IdentifierSchemes.CRSID),
                set(),
                None,
                None,
            ),
        )

    def test_returns_authenticated_non_anonymous_user(self):
        user, _ = self.auth.authenticate(
            self.request_with_headers(
                {
                    "x-api-org-name": "test",
                    "x-api-developer-app-class": "public",
                    "x-api-oauth2-user": str(Identifier("a123", IdentifierSchemes.CRSID)),
                }
            )
        )
        self.assertFalse(user.is_anonymous)
        self.assertTrue(user.is_authenticated)

    def test_will_pass_through_scopes(self):
        _, auth = self.auth.authenticate(
            self.request_with_headers(
                {
                    "x-api-org-name": "test",
                    "x-api-developer-app-class": "public",
                    "x-api-oauth2-user": str(Identifier("a123", IdentifierSchemes.CRSID)),
                    "x-api-oauth2-scope": (
                        "https://api.apps.cam.ac.uk/a.readonly https://api.apps.cam.ac.uk/b"
                    ),
                }
            )
        )

        self.assertEqual(
            auth,
            APIGatewayAuthenticationDetails(
                Identifier("a123", IdentifierSchemes.CRSID),
                set(
                    [
                        "https://api.apps.cam.ac.uk/a.readonly",
                        "https://api.apps.cam.ac.uk/b",
                    ]
                ),
                None,
                None,
            ),
        )

    def test_will_pass_through_app_and_client_ids(self):
        _, auth = self.auth.authenticate(
            self.request_with_headers(
                {
                    "x-api-org-name": "test",
                    "x-api-developer-app-class": "confidential",
                    "x-api-oauth2-user": str(Identifier("a123", IdentifierSchemes.CRSID)),
                    "x-api-oauth2-scope": (
                        "https://api.apps.cam.ac.uk/a.readonly https://api.apps.cam.ac.uk/b"
                    ),
                    "x-api-developer-app-id": "app-uuid-mock",
                    "x-api-oauth2-client-id": "client-id-uuid-mock",
                }
            )
        )

        self.assertEqual(
            auth,
            APIGatewayAuthenticationDetails(
                Identifier("a123", IdentifierSchemes.CRSID),
                set(
                    [
                        "https://api.apps.cam.ac.uk/a.readonly",
                        "https://api.apps.cam.ac.uk/b",
                    ]
                ),
                "app-uuid-mock",
                "client-id-uuid-mock",
            ),
        )

    def test_fails_authentication_if_no_header_present(self):
        with self.assertRaisesMessage(AuthenticationFailed, "Bearer token not present"):
            self.auth.authenticate(
                self.request_with_headers(
                    {
                        "x-api-org-name": "test",
                        "x-api-developer-app-class": "public",
                        "x-api-oauth2-user": str(Identifier("a123", IdentifierSchemes.CRSID)),
                    },
                    include_authorization=False,
                )
            )

    def test_fails_with_invalid_token(self):
        self.mock_verify_token.side_effect = GoogleAuthError("token-failed-verification")
        with self.assertRaisesMessage(AuthenticationFailed, "Invalid API Gateway token"):
            self.auth.authenticate(
                self.request_with_headers(
                    {
                        "x-api-org-name": "test",
                        "x-api-developer-app-class": "public",
                        "x-api-oauth2-user": str(Identifier("a123", IdentifierSchemes.CRSID)),
                    }
                )
            )

    @override_settings(
        API_GATEWAY_JWT_EXPECTED_AUDIENCE="https://audience.invalid/",
        API_GATEWAY_JWT_ISSUER_CERTS_URL="https://issuer.invalid/certs",
    )
    def test_verify_token_passed_expected_audience_and_certs_url(self):
        self.assertIsNotNone(
            self.auth.authenticate(
                self.request_with_headers(
                    {
                        "x-api-org-name": "test",
                        "x-api-developer-app-class": "public",
                        "x-api-oauth2-user": str(Identifier("a123", IdentifierSchemes.CRSID)),
                    }
                )
            )
        )
        self.mock_verify_token.assert_called_once()
        self.assertEqual(self.mock_verify_token.call_args[0][0], self.expected_token)
        self.assertEqual(
            self.mock_verify_token.call_args[1]["audience"], "https://audience.invalid/"
        )
        self.assertEqual(
            self.mock_verify_token.call_args[1]["certs_url"], "https://issuer.invalid/certs"
        )

    def test_bad_id_token(self):
        with self.assertRaisesMessage(AuthenticationFailed, "Invalid API Gateway token"):
            self.auth.authenticate(
                self.request_with_headers(
                    {
                        "x-api-org-name": "test",
                        "x-api-developer-app-class": "public",
                        "x-api-oauth2-user": str(Identifier("a123", IdentifierSchemes.CRSID)),
                        "Authorization": f"Bearer {secrets.token_urlsafe()}",
                    },
                    include_authorization=False,
                )
            )

    @override_settings(API_GATEWAY_JWT_TRUSTED_ISSUERS=["https://issuer.invalid/"])
    def test_bad_id_token_issuer(self):
        with self.assertRaisesMessage(AuthenticationFailed, "Invalid API Gateway token"):
            self.auth.authenticate(
                self.request_with_headers(
                    {
                        "x-api-org-name": "test",
                        "x-api-developer-app-class": "public",
                        "x-api-oauth2-user": str(Identifier("a123", IdentifierSchemes.CRSID)),
                    }
                )
            )

    @override_settings(API_GATEWAY_JWT_EXPECTED_AUTHORISED_PARTIES=["gateway@gateway.invalid"])
    def test_bad_azp_claim(self):
        with self.assertRaisesMessage(AuthenticationFailed, "Invalid API Gateway token"):
            self.auth.authenticate(
                self.request_with_headers(
                    {
                        "x-api-org-name": "test",
                        "x-api-developer-app-class": "public",
                        "x-api-oauth2-user": str(Identifier("a123", IdentifierSchemes.CRSID)),
                    }
                )
            )



# This import needs to be removed but to do so breaks compatibility with
# existing users. See comments in apigatewayauth.auth.APIGatewayAuthentication.
from .authentication import (  # noqa: F401
    APIGatewayAuthentication,
    APIGatewayAuthenticationDetails,
)

default_app_config = "apigatewayauth.apps.APIGatewayAuthConfig"



"""
Utilities for verifying id tokens.
"""
from typing import Any, Iterable, Optional

import cachecontrol
import requests
from google.auth.exceptions import GoogleAuthError
from google.auth.transport import Request
from google.auth.transport.requests import Request as RequestsRequest
from google.oauth2 import id_token

__all__ = ["verify_id_token_for_api_backend", "InvalidIdTokenError"]


def _default_caching_request() -> Request:
    """
    Return a google.auth.transport.Request sub-class which caches requests for id token
    certificates.

    Uses the approach documented at
    https://google-auth.readthedocs.io/en/latest/reference/google.oauth2.id_token.html.

    """
    session = requests.session()
    cached_session = cachecontrol.CacheControl(session)
    return RequestsRequest(session=cached_session)


# Default HTTP transport used to fetch public keys for verifying tokens.
_DEFAULT_CACHING_REQUEST = _default_caching_request()

# List of Google service accounts which represent the API gateway identity.
_DEFAULT_EXPECTED_AUTHORISED_PARTIES = ["api-gateway@api-meta-2555105a.iam.gserviceaccount.com"]

# Default URL used to fetch certificates corresponding to JWT signing keys
_DEFAULT_CERTS_URL = "https://www.googleapis.com/oauth2/v1/certs"

# Default list of issuers trusted to issue identity tokens.
_DEFAULT_TRUSTED_ISSUERS = {"accounts.google.com", "https://accounts.google.com"}


class InvalidIdTokenError(ValueError):
    """
    Exception raised if token verification fails. The exception message describes the reason that
    the token failed verification.
    """


def verify_id_token_for_api_backend(
    token: str,
    expected_audience: str,
    *,
    request: Optional[Request] = None,
    expected_authorised_parties: Optional[Iterable[str]] = None,
    certs_url: Optional[str] = None,
    trusted_issuers: Optional[Iterable[str]] = None,
) -> dict[str, Any]:
    """
    Verify an incoming authentication token from the Gateway sent to an API backend. Use this
    function when implementing API backends.

    :param token: id token passed to the backend by the Gateway.
    :param expected_audience: expected audience for id token. For Cloud Run-hosted backends this is
        the base URL of the application.
    :param request: HTTP transport implementation. If `None` then a default implementation based on
        the :py:mod:`requests` library is used. Note that certificates are fetched on each call to
        this function and so the transport implementation should support caching. This is the case
        for the default implementation.
    :param expected_authorised_parties: the "azp" claim in the id token must match one of the
        strings in this iterable for the token to be verified. If `None` then a list of known API
        Gateway identities is used.
    :param certs_url: URL used to fetch certificates. If omitted, a URL corresponding to Google
        identity tokens is used.
    :param trusted_issuers: iterable of strings specifying trusted id token issuers. If omitted, a
        set of issuers corresponding to Google is used.
    :returns: a dictionary containing the id token payload.
    """
    request = request if request is not None else _DEFAULT_CACHING_REQUEST
    expected_authorised_parties = (
        set(expected_authorised_parties)
        if expected_authorised_parties is not None
        else _DEFAULT_EXPECTED_AUTHORISED_PARTIES
    )
    certs_url = certs_url if certs_url is not None else _DEFAULT_CERTS_URL
    trusted_issuers = (
        set(trusted_issuers) if trusted_issuers is not None else _DEFAULT_TRUSTED_ISSUERS
    )

    try:
        payload = id_token.verify_token(
            token, request, audience=expected_audience, certs_url=certs_url
        )
    except GoogleAuthError as e:
        raise InvalidIdTokenError(str(e)) from e
    if payload["iss"] not in trusted_issuers:
        raise InvalidIdTokenError(f"'{payload['iss']}' is not a trusted id token issuer")

    authorized_party = payload.get("azp", "")
    if payload.get("azp", "") not in expected_authorised_parties:
        raise InvalidIdTokenError(
            f"Authorized party {authorized_party!r} not one of {expected_authorised_parties!r}."
        )

    return payload
